#version 430 core
out float FragColor;

in vec2 ViewRay;

uniform sampler2D gPosition;
uniform sampler2D gNormal;
uniform sampler2D texNoise;

uniform vec3 samples[64];

uniform int sampleKernelSize = 16;
uniform float radius = 2.0;	
uniform float bias = 0.001;
uniform float strength = 1.5;

uniform mat4 projection;

mat3 GetRotationMatrix(vec3 surfaceNormal, vec3 randomNormal)
{
	// create TBN change-of-basis matrix: from tangent-space to view-space
	// Gram–Schmidt process to construct an orthogonal basis.
	
	// re-orthogonalize T with respect to N
	vec3 tangent = normalize(randomNormal - surfaceNormal * dot(randomNormal, surfaceNormal));
	// retrieve perpendicular vector B with the cross product of T and N
	vec3 bitangent = cross(surfaceNormal, tangent);
	return mat3(tangent, bitangent, surfaceNormal);
}

void main()
{
	vec2 texSize  = textureSize(gPosition, 0).xy;
	vec2 TexCoords = gl_FragCoord.xy / texSize;
	vec3 currentPos = texture(gPosition, TexCoords).xyz;

	vec2 noiseScale = texSize/8;
	
    vec3 normal = normalize(texture(gNormal, TexCoords).rgb);
	
	//minimize banding effect by randomly rotating the sample kernel
    vec3 randomVec = texture(texNoise, TexCoords * noiseScale).xyz; 
	
    mat3 TBN = GetRotationMatrix(normal, randomVec);
	
	int kernelSize = min(sampleKernelSize,64);
    
    float occlusion = 0.0;
	//iterate over the sample kernel and calculate occlusion factor
    for(int i = 0; i < kernelSize; ++i)
    {
        vec3 samplePosDir = TBN * samples[i]; // from unit hemisphere to view-space
        vec3 samplePos = currentPos + samplePosDir * radius; 
        
        // project sample to screen
        vec4 offset = vec4(samplePos, 1.0);
        offset = projection * offset; // from view to clip-space
        vec2 sampleTex = (offset.xy/offset.w) * 0.5 + 0.5; // transform to range 0.0 - 1.0
        
        // get sample depth
        float sampleDepth = texture(gPosition, sampleTex).z; // get depth value of kernel sample
		        
        // range check & accumulate
		if ((abs(currentPos.z - sampleDepth) < radius) && 
			(sampleDepth >= samplePos.z + bias)) {
			 //if sampleDepth is inside of geometry, the area gets darker
			occlusion += 1.0;
		}
		                
    }
	
    occlusion = 1.0 - (occlusion / kernelSize);
    
    FragColor = pow(occlusion, strength);

}