#version 430 core

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
} vert;

layout(location = 0) out vec3 color;

uniform vec3 camera_world;

uniform sampler2D albedoMap;
uniform sampler2D metallicRoughnessMap;
uniform sampler2D ssaoMap;
uniform sampler2D normalMap;
uniform samplerCube depthMap[4];

uniform int shadowMode;
uniform int aoOn;

uniform vec4 pbrTextures;
uniform vec2 near_far_light;


uniform float sh_bias;
uniform int sh_samples;

uniform struct PointLight {
	vec3 color;
	vec3 position;
};

#define MAX_NR_OF_LIGHTS 99
uniform PointLight pointLights[MAX_NR_OF_LIGHTS];
uniform int numberOfLights;

const float PI = 3.14159265359;


vec3 getNormalFromMap()
{
    vec3 tangentNormal = texture(normalMap, vert.uv).xyz * 2.0 - 1.0;

    vec3 Q1  = dFdx(vert.position_world);
    vec3 Q2  = dFdy(vert.position_world);
    vec2 st1 = dFdx(vert.uv);
    vec2 st2 = dFdy(vert.uv);

    vec3 N   = normalize(vert.normal_world);
    vec3 T  = normalize(Q1*st2.t - Q2*st1.t);
    vec3 B  = -normalize(cross(N, T));
    mat3 TBN = mat3(T, B, N);

    return normalize(TBN * tangentNormal);
}

float DistributionGGX(float NdotH, float roughness)
{
    float a = roughness*roughness;
    float a2 = a*a;
    float NdotH2 = NdotH*NdotH;

    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;

    return a2 / max(denom, 0.0000001);
}


float GeometrySmith(float NdotV, float NdotL, float roughness)
{
	float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float ggx2 = NdotV / (NdotV * (1.0 - k) + k);
    float ggx1 = NdotL / (NdotL * (1.0 - k) + k);

    return ggx1 * ggx2;
}

vec3 fresnelSchlick(float HdotV, vec3 baseReflectivity)
{
    return baseReflectivity + (1.0 - baseReflectivity) * pow(1.0 - HdotV, 5.0);
}

//PCF fixed samples
vec3 sampleOffsetDirections[20] = vec3[]
(
   vec3( 1,  1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1,  1,  1), 
   vec3( 1,  1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1,  1, -1),
   vec3( 1,  1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1,  1,  0),
   vec3( 1,  0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1,  0, -1),
   vec3( 0,  1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0,  1, -1)
);   

float ShadowCalculation(vec3 fragPos, vec3 lightPosition, int j)
{
    // get vector between fragment position and light position
    vec3 fragToLight = fragPos - lightPosition;	
	
	//distance light to fragment
    float currentDepth = length(fragToLight);
	
	//PCF algorithm
	float shadow = 0.0f;
	
	if (shadowMode==0) {
		float viewDistance = length(camera_world - fragPos);
		//float diskRadius = (1.0 + (viewDistance / near_far_light.y)) / 20.0; 
		float diskRadius = 0.03; 
	
		for(int i = 0; i < sh_samples; ++i)
		{
			float closestDepth = texture(depthMap[j], fragToLight + sampleOffsetDirections[i] * diskRadius).r;
			closestDepth = closestDepth * (near_far_light.y - near_far_light.x) + near_far_light.x;

			if((closestDepth +  sh_bias) <  currentDepth) {
			shadow += 1.0;
			}
		}
	
		shadow = shadow / float(sh_samples);
		
	} else {
	
		float closestDepth = texture(depthMap[j], fragToLight).r;
		closestDepth = closestDepth * (near_far_light.y - near_far_light.x) + near_far_light.x;  
		
		if((closestDepth +  sh_bias) <  currentDepth) {
			shadow = 1.0;
		} else {
			shadow = 0;
		}
	}
	

    return shadow;
	  
}


void main() {	
	
	// Fill pbr values according to textures. Maybe use gltf factors as base values?
	vec3 albedo = vec3(1,0,1);
	if (pbrTextures.r == 1) {
		albedo = pow(texture(albedoMap, vert.uv).rgb, vec3(2.2));
	}
	float metallic = 0;
	float roughness = 0.5;
	if (pbrTextures.g == 1) {
		metallic = texture(metallicRoughnessMap, vert.uv).b;
		roughness = texture(metallicRoughnessMap, vert.uv).g;
	}
	
	vec3 N = normalize(vert.normal_world);
	if (pbrTextures.a == 1) {
		N = getNormalFromMap();
	}
	
	vec2 ssaoBlurTexSize  = textureSize(ssaoMap, 0).xy;
	vec2 ssaoBlurTexCoord = (gl_FragCoord.xy/2) / ssaoBlurTexSize;
	
	float ssao = texture(ssaoMap, ssaoBlurTexCoord).r;
	
    vec3 V = normalize(camera_world - vert.position_world);

	vec3 baseReflectivity = vec3(0.04); 
    baseReflectivity = mix(baseReflectivity, albedo, metallic);

	vec3 Lo = vec3(0.0);
	float shadow = 0.0;
	
	// Iterate over lights array, only for indices with actual lights
	for (int i = 0; i < numberOfLights; i++)
	{
		vec3 L = normalize(pointLights[i].position - vert.position_world);
		vec3 H = normalize(V + L);
		float distance = length(pointLights[i].position - vert.position_world);
		float attenuation = 1.0 / (distance * distance);
		vec3 radiance = (pointLights[i].color * 0.3) * attenuation;

		float NdotV = max(dot(N, V), 0.0000001);
		float NdotL = max(dot(N, L), 0.0000001);
		float HdotV = max(dot(H, V), 0.0);
		float NdotH = max(dot(N, H), 0.0);

		float NDF = DistributionGGX(NdotH, max(roughness, 0.05));
		float G   = GeometrySmith(NdotV, NdotL, max(roughness, 0.05));
		vec3 F    = fresnelSchlick(HdotV, baseReflectivity);

		vec3 nominator    = NDF * G * F; 
		float denominator = 4 * NdotV * NdotL;
		vec3 specular = nominator / denominator;

		vec3 kS = F;

		vec3 kD = vec3(1.0) - kS;
		kD *= 1.0 - metallic;	  

		if(shadowMode==2) {
			Lo += (kD * albedo / PI + specular) * radiance * NdotL;
		} else {
			shadow = ShadowCalculation(vert.position_world,pointLights[i].position, i);
			Lo += (kD * albedo / PI + specular) * radiance * NdotL * (1.0 - shadow);
			
		}	
		
	}
	
	vec3 ambient = vec3(0.5) * albedo ;

	if(aoOn==1) {
		color = ambient  + Lo*ssao;
	}	else {
		color = ambient + Lo;
	}

	// HDR tonemapping
	color = color / (color + vec3(1.0));
	// gamma correct
	color = pow(color, vec3(1.0/2.2)); 

}

